@@ -22,7 +22,7 @@
  * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
-/* $Id: softflowd.c,v 1.80 2005/01/10 01:50:07 djm Exp $ */
+/* $Id: softflowd.c,v 1.93 2006/11/02 06:23:29 djm Exp $ */
 
 /*
  * This is software implementation of Cisco's NetFlow(tm) traffic 
@@ -51,7 +51,7 @@
 #include "log.h"
 #include <pcap.h>
 
-RCSID("$Id: softflowd.c,v 1.80 2005/01/10 01:50:07 djm Exp $");
+RCSID("$Id: softflowd.c,v 1.93 2006/11/02 06:23:29 djm Exp $");
 
 /* Global variables */
 static int verbose_flag = 0;		/* Debugging flag */
@@ -83,7 +83,10 @@ struct DATALINK {
 static const struct DATALINK lt[] = {
 	{ DLT_EN10MB,	14, 12,  2,  1, 0xffffffff,  0x0800,   0x86dd },
 	{ DLT_PPP,	 5,  3,  2,  1, 0xffffffff,  0x0021,   0x0057 },
-	{ DLT_RAW,	 0,  0,  1,  1, 0x000000f0,  0x0004,   0x0006 },
+#ifdef DLT_LINUX_SLL
+	{ DLT_LINUX_SLL,16, 14,  2,  1, 0xffffffff,  0x0800,   0x86dd },
+#endif
+	{ DLT_RAW,	 0,  0,  1,  1, 0x000000f0,  0x0040,   0x0060 },
 	{ DLT_NULL,	 4,  0,  4,  0, 0xffffffff, AF_INET, AF_INET6 },
 #ifdef DLT_LOOP
 	{ DLT_LOOP,	 4,  0,  4,  1, 0xffffffff, AF_INET, AF_INET6 },
@@ -92,7 +95,7 @@ static const struct DATALINK lt[] = {
 };
 
 /* Netflow send functions */
-typedef int (netflow_send_func_t)(struct FLOW **, int, int, u_int64_t,
+typedef int (netflow_send_func_t)(struct FLOW **, int, int, u_int64_t *,
     struct timeval *, int);
 struct NETFLOW_SENDER {
 	int version;
@@ -249,7 +252,7 @@ format_flow(struct FLOW *flow)
 	    flow->octets[0], flow->packets[0], 
 	    flow->octets[1], flow->packets[1], 
 	    stime, (flow->flow_start.tv_usec + 500) / 1000, 
-	    ftime, (flow->flow_start.tv_usec + 500) / 1000,
+	    ftime, (flow->flow_last.tv_usec + 500) / 1000,
 	    flow->tcp_flags[0], flow->tcp_flags[1],
 	    flow->ip6_flowlabel[0], flow->ip6_flowlabel[1]);
 
@@ -282,6 +285,7 @@ transport_to_flowrec(struct FLOW *flow, const u_int8_t *pkt,
 {
 	const struct tcphdr *tcp = (const struct tcphdr *)pkt;
 	const struct udphdr *udp = (const struct udphdr *)pkt;
+	const struct icmp *icmp = (const struct icmp *)pkt;
 
 	/*
 	 * XXX to keep flow in proper canonical format, it may be necessary
@@ -306,6 +310,15 @@ transport_to_flowrec(struct FLOW *flow, const u_int8_t *pkt,
 		flow->port[ndx] = udp->uh_sport;
 		flow->port[ndx ^ 1] = udp->uh_dport;
 		break;
+	case IPPROTO_ICMP:
+		/*
+		 * Encode ICMP type * 256 + code into dest port like
+		 * Cisco routers
+		 */
+		flow->port[ndx] = 0;
+		flow->port[ndx ^ 1] = htons(icmp->icmp_type * 256 +
+		    icmp->icmp_code);
+		break;
 	}
 	return (0);
 }
@@ -470,7 +483,7 @@ flow_update_expiry(struct FLOWTRACK *ft, struct FLOW *flow)
 	if (ft->icmp_timeout != 0 &&
 	    ((flow->af == AF_INET && flow->protocol == IPPROTO_ICMP) || 
 	    ((flow->af == AF_INET6 && flow->protocol == IPPROTO_ICMPV6)))) {
-		/* UDP flows */
+		/* ICMP flows */
 		flow->expiry->expires_at = flow->flow_last.tv_sec + 
 		    ft->icmp_timeout;
 		flow->expiry->reason = R_ICMP;
@@ -483,6 +496,11 @@ flow_update_expiry(struct FLOWTRACK *ft, struct FLOW *flow)
 	flow->expiry->reason = R_GENERAL;
 
  out:
+	if (ft->maximum_lifetime != 0 && flow->expiry->expires_at != 0) {
+		flow->expiry->expires_at = MIN(flow->expiry->expires_at,
+		    flow->flow_start.tv_sec + ft->maximum_lifetime);
+	}
+
 	EXPIRY_INSERT(EXPIRIES, &ft->expiries, flow->expiry);
 }
 
@@ -530,11 +548,27 @@ process_packet(struct FLOWTRACK *ft, const u_int8_t *pkt, int af,
 	if (frag)
 		ft->frag_packets++;
 
+	/* Zero out bits of the flow that aren't relevant to tracking level */
+	switch (ft->track_level) {
+	case TRACK_IP_ONLY:
+		tmp.protocol = 0;
+		/* FALLTHROUGH */
+	case TRACK_IP_PROTO:
+		tmp.port[0] = tmp.port[1] = 0;
+		tmp.tcp_flags[0] = tmp.tcp_flags[1] = 0;
+		/* FALLTHROUGH */
+	case TRACK_FULL:
+		break;
+	}
+
 	/* If a matching flow does not exist, create and insert one */
 	if ((flow = FLOW_FIND(FLOWS, &ft->flows, &tmp)) == NULL) {
 		/* Allocate and fill in the flow */
-		if ((flow = malloc(sizeof(*flow))) == NULL)
+		if ((flow = malloc(sizeof(*flow))) == NULL) {
+			logit(LOG_ERR, "process_packet: flow malloc(%u) fail",
+			    sizeof(*flow));
 			return (PP_MALLOC_FAIL);
+		}
 		memcpy(flow, &tmp, sizeof(*flow));
 		memcpy(&flow->flow_start, received_time,
 		    sizeof(flow->flow_start));
@@ -542,8 +576,11 @@ process_packet(struct FLOWTRACK *ft, const u_int8_t *pkt, int af,
 		FLOW_INSERT(FLOWS, &ft->flows, flow);
 
 		/* Allocate and fill in the associated expiry event */
-		if ((flow->expiry = malloc(sizeof(*flow->expiry))) == NULL)
+		if ((flow->expiry = malloc(sizeof(*flow->expiry))) == NULL) {
+			logit(LOG_ERR, "process_packet: expiry malloc(%u) fail",
+			    sizeof(*flow->expiry));
 			return (PP_MALLOC_FAIL);
+		}
 		flow->expiry->flow = flow;
 		/* Must be non-zero (0 means expire immediately) */
 		flow->expiry->expires_at = 1;
@@ -736,9 +773,18 @@ check_expired(struct FLOWTRACK *ft, struct NETFLOW_TARGET *target, int ex)
 		    (ex != CE_EXPIRE_FORCED &&
 		    (expiry->expires_at < now.tv_sec))) {
 			/* Flow has expired */
+
+			if (ft->maximum_lifetime != 0 && 
+	    		    expiry->flow->flow_last.tv_sec - 
+			    expiry->flow->flow_start.tv_sec >= 
+	    		    ft->maximum_lifetime)
+					expiry->reason = R_MAXLIFE;
+
 			if (verbose_flag)
-				logit(LOG_DEBUG, "Queuing flow seq:%llu (%p) for expiry",
-				   expiry->flow->flow_seq, expiry->flow);
+				logit(LOG_DEBUG,
+				    "Queuing flow seq:%llu (%p) for expiry "
+				    "reason %d", expiry->flow->flow_seq,
+				    expiry->flow, expiry->reason);
 
 			/* Add to array of expired flows */
 			oldexp = expired_flows;
@@ -774,12 +820,11 @@ check_expired(struct FLOWTRACK *ft, struct NETFLOW_TARGET *target, int ex)
 	if (num_expired > 0) {
 		if (target != NULL && target->fd != -1) {
 			r = target->dialect->func(expired_flows, num_expired, 
-			    target->fd, ft->flows_exported, 
+			    target->fd, &ft->flows_exported, 
 			    &ft->system_boot_time,  verbose_flag);
 			if (verbose_flag)
 				logit(LOG_DEBUG, "sent %d netflow packets", r);
 			if (r > 0) {
-				ft->flows_exported += num_expired * 2;
 				ft->packets_sent += r;
 				/* XXX what if r < num_expired * 2 ? */
 			} else {
@@ -825,7 +870,7 @@ force_expire(struct FLOWTRACK *ft, u_int32_t num_to_expire)
 	 * alter them 
 	 */
 	 
-	if ((expiryv = malloc(sizeof(*expiryv) * num_to_expire)) == NULL) {
+	if ((expiryv = calloc(num_to_expire, sizeof(*expiryv))) == NULL) {
 		/*
 		 * On malloc failure, expire ALL flows. I assume that 
 		 * setting all the keys in a tree to the same value is 
@@ -894,11 +939,12 @@ delete_all_flows(struct FLOWTRACK *ft)
  * and the tree of expiry events.
  */
 static int
-statistics(struct FLOWTRACK *ft, FILE *out)
+statistics(struct FLOWTRACK *ft, FILE *out, pcap_t *pcap)
 {
 	int i;
 	struct protoent *pe;
 	char proto[32];
+	struct pcap_stat ps;
 
 	fprintf(out, "Number of active flows: %d\n", ft->num_flows);
 	fprintf(out, "Packets processed: %llu\n", ft->total_packets);
@@ -910,6 +956,15 @@ statistics(struct FLOWTRACK *ft, FILE *out)
 	fprintf(out, "Flows exported: %llu in %llu packets (%llu failures)\n",
 	    ft->flows_exported, ft->packets_sent, ft->flows_dropped);
 
+	if (pcap_stats(pcap, &ps) == 0) {
+		fprintf(out, "Packets received by libpcap: %lu\n",
+		    (unsigned long)ps.ps_recv);
+		fprintf(out, "Packets dropped by libpcap: %lu\n",
+		    (unsigned long)ps.ps_drop);
+		fprintf(out, "Packets dropped by interface: %lu\n",
+		    (unsigned long)ps.ps_ifdrop);
+	}
+
 	fprintf(out, "\n");
 
 	if (ft->flows_expired != 0) {
@@ -1069,7 +1124,7 @@ print_timeouts(struct FLOWTRACK *ft, FILE *out)
 
 static int
 accept_control(int lsock, struct NETFLOW_TARGET *target, struct FLOWTRACK *ft,
-    int *exit_request, int *stop_collection_flag)
+    pcap_t *pcap, int *exit_request, int *stop_collection_flag)
 {
 	unsigned char buf[64], *p;
 	FILE *ctlf;
@@ -1102,8 +1157,11 @@ accept_control(int lsock, struct NETFLOW_TARGET *target, struct FLOWTRACK *ft,
 	ret = -1;
 	if (strcmp(buf, "help") == 0) {
 		fprintf(ctlf, "Valid control words are:\n");
-		fprintf(ctlf, "\tdebug+ debug- delete-all dump-flows exit expire-all\n");
-		fprintf(ctlf, "\tshutdown start-gather statistics stop-gather timeouts\n");
+		fprintf(ctlf, "\tdebug+ debug- delete-all dump-flows exit "
+		    "expire-all\n");
+		fprintf(ctlf, "\tshutdown start-gather statistics stop-gather "
+		    "timeouts\n");
+		fprintf(ctlf, "\tsend-template\n");
 		ret = 0;
 	} else if (strcmp(buf, "shutdown") == 0) {
 		fprintf(ctlf, "softflowd[%u]: Shutting down gracefully...\n", getpid());
@@ -1114,9 +1172,15 @@ accept_control(int lsock, struct NETFLOW_TARGET *target, struct FLOWTRACK *ft,
 		*exit_request = 1;
 		ret = 1;
 	} else if (strcmp(buf, "expire-all") == 0) {
+		netflow9_resend_template();
 		fprintf(ctlf, "softflowd[%u]: Expired %d flows.\n", getpid(), 
 		    check_expired(ft, target, CE_EXPIRE_ALL));
 		ret = 0;
+	} else if (strcmp(buf, "send-template") == 0) {
+		netflow9_resend_template();
+		fprintf(ctlf, "softflowd[%u]: Template will be sent at "
+		    "next flow export\n", getpid());
+		ret = 0;
 	} else if (strcmp(buf, "delete-all") == 0) {
 		fprintf(ctlf, "softflowd[%u]: Deleted %d flows.\n", getpid(), 
 		    delete_all_flows(ft));
@@ -1124,7 +1188,7 @@ accept_control(int lsock, struct NETFLOW_TARGET *target, struct FLOWTRACK *ft,
 	} else if (strcmp(buf, "statistics") == 0) {
 		fprintf(ctlf, "softflowd[%u]: Accumulated statistics:\n", 
 		    getpid());
-		statistics(ft, ctlf);
+		statistics(ft, ctlf, pcap);
 		ret = 0;
 	} else if (strcmp(buf, "debug+") == 0) {
 		fprintf(ctlf, "softflowd[%u]: Debug level increased.\n",
@@ -1331,6 +1395,8 @@ init_flowtrack(struct FLOWTRACK *ft)
 	FLOW_INIT(&ft->flows);
 	EXPIRY_INIT(&ft->expiries);
 	
+	ft->track_level = TRACK_FULL;
+
 	ft->tcp_timeout = DEFAULT_TCP_TIMEOUT;
 	ft->tcp_rst_timeout = DEFAULT_TCP_RST_TIMEOUT;
 	ft->tcp_fin_timeout = DEFAULT_TCP_FIN_TIMEOUT;
@@ -1375,20 +1441,21 @@ usage(void)
 {
 	fprintf(stderr, "Usage: %s [options] [bpf_program]\n", PROGNAME);
 	fprintf(stderr, "This is %s version %s. Valid commandline options:\n", PROGNAME, PROGVER);
-	fprintf(stderr, "  -i interface    Specify interface to listen on\n");
-	fprintf(stderr, "  -r pcap_file    Specify packet capture file to read\n");
-	fprintf(stderr, "  -t timeout=time Specify named timeout\n");
-	fprintf(stderr, "  -m max_flows    Specify maximum number of flows to track (default %d)\n", DEFAULT_MAX_FLOWS);
-	fprintf(stderr, "  -n host:port    Send Cisco NetFlow(tm)-compatible packets to host:port\n");
-	fprintf(stderr, "  -p pidfile      Record pid in specified file (default: %s)\n", DEFAULT_PIDFILE);
-	fprintf(stderr, "  -c pidfile      Location of control socket (default: %s)\n", DEFAULT_CTLSOCK);
-	fprintf(stderr, "  -v 1|5|9        NetFlow export packet version\n");
-	fprintf(stderr, "  -L hoplimit     Set TTL/hoplimit for export datagrams\n");
-	fprintf(stderr, "  -6              Track IPv6 flows, regardless of whether selected \n"
-	                "                  NetFlow export protocol supports it\n");
-	fprintf(stderr, "  -d              Don't daemonise\n");
-	fprintf(stderr, "  -D              Debug mode: don't daemonise + verbosity + track v6 flows\n");
-	fprintf(stderr, "  -h              Display this help\n");
+	fprintf(stderr, "  -i interface     Specify interface to listen on\n");
+	fprintf(stderr, "  -r pcap_file     Specify packet capture file to read\n");
+	fprintf(stderr, "  -t timeout=time  Specify named timeout\n");
+	fprintf(stderr, "  -m max_flows     Specify maximum number of flows to track (default %d)\n", DEFAULT_MAX_FLOWS);
+	fprintf(stderr, "  -n host:port     Send Cisco NetFlow(tm)-compatible packets to host:port\n");
+	fprintf(stderr, "  -p pidfile       Record pid in specified file (default: %s)\n", DEFAULT_PIDFILE);
+	fprintf(stderr, "  -c pidfile       Location of control socket (default: %s)\n", DEFAULT_CTLSOCK);
+	fprintf(stderr, "  -v 1|5|9         NetFlow export packet version\n");
+	fprintf(stderr, "  -L hoplimit      Set TTL/hoplimit for export datagrams\n");
+	fprintf(stderr, "  -T full|proto|ip Set flow tracking level (default: full)\n");
+	fprintf(stderr, "  -6               Track IPv6 flows, regardless of whether selected \n"
+	                "                   NetFlow export protocol supports it\n");
+	fprintf(stderr, "  -d               Don't daemonise\n");
+	fprintf(stderr, "  -D               Debug mode: don't daemonise + verbosity + track v6 flows\n");
+	fprintf(stderr, "  -h               Display this help\n");
 	fprintf(stderr, "\n");
 	fprintf(stderr, "Valid timeout names and default values:\n");
 	fprintf(stderr, "  tcp     (default %6d)", DEFAULT_TCP_TIMEOUT);
@@ -1593,7 +1660,7 @@ main(int argc, char **argv)
 	ctlsock_path = DEFAULT_CTLSOCK;
 	dontfork_flag = 0;
 	always_v6 = 0;
-	while ((ch = getopt(argc, argv, "6hdDL:i:r:f:t:n:m:p:c:v:")) != -1) {
+	while ((ch = getopt(argc, argv, "6hdDL:T:i:r:f:t:n:m:p:c:v:")) != -1) {
 		switch (ch) {
 		case '6':
 			always_v6 = 1;
@@ -1630,6 +1697,19 @@ main(int argc, char **argv)
 			/* Will exit on failure */
 			set_timeout(&flowtrack, optarg); 
 			break;
+		case 'T':
+			if (strcasecmp(optarg, "full") == 0)
+				flowtrack.track_level = TRACK_FULL;
+			else if (strcasecmp(optarg, "proto") == 0)
+				flowtrack.track_level = TRACK_IP_PROTO;
+			else if (strcasecmp(optarg, "ip") == 0)
+				flowtrack.track_level = TRACK_IP_ONLY;
+			else {
+				fprintf(stderr, "Unknown flow tracking level\n");
+				usage();
+				exit(1);
+			}
+			break;
 		case 'L':
 			hoplimit = atoi(optarg);
 			if (hoplimit < 0 || hoplimit > 255) {
@@ -1740,6 +1820,7 @@ main(int argc, char **argv)
 	/* Main processing loop */
 	gettimeofday(&flowtrack.system_boot_time, NULL);
 	stop_collection_flag = 0;
+	memset(&cb_ctxt, '\0', sizeof(cb_ctxt));
 	cb_ctxt.ft = &flowtrack;
 	cb_ctxt.linktype = linktype;
 	cb_ctxt.want_v6 = target.dialect->v6_capable || always_v6;
@@ -1773,7 +1854,7 @@ main(int argc, char **argv)
 
 		/* Accept connection on control socket if present */
 		if (ctlsock != -1 && pl[1].revents != 0) {
-			if (accept_control(ctlsock, &target, &flowtrack, 
+			if (accept_control(ctlsock, &target, &flowtrack, pcap,
 			    &exit_request, &stop_collection_flag) != 0)
 				break;
 		}
@@ -1838,7 +1919,7 @@ expiry_check:
 		logit(LOG_ERR, "Exiting immediately on internal error");
 		
 	if (capfile != NULL && dontfork_flag)
-		statistics(&flowtrack, stdout);
+		statistics(&flowtrack, stdout, pcap);
 
 	pcap_close(pcap);
 	
